/* -*- mode: c++ -*-
   This program is part of the LifeV library
   Copyright (C) 2001,2002,2003,2004 EPFL, INRIA, Politechnico di Milano

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <life/lifesolver/FSIOperator.hpp>
#include <life/lifesolver/reducedLinFluid.hpp>

namespace LifeV
{
// Constructors


// Destructor
FSIOperator::~FSIOperator()
{
}

//
void
FSIOperator::setup()
{
    if ( M_solid && M_fluid )
    {
        M_dispStruct.resize( 3*M_solid->dDof().numTotalDof() );
        M_velo.resize( 3*M_solid->dDof().numTotalDof() );

        M_dofFluidToStructure->setup(M_solid->mesh().getRefFE(), M_solid->dDof(),
                                     M_fluid->mesh().getRefFE(), M_fluid->uDof());
        M_dofFluidToStructure->update(M_solid->mesh(), 1,
                                      M_fluid->mesh(), 1,
                                      0.);

        M_dofStructureToSolid->setup(M_solid->mesh().getRefFE(), M_solid->dDof(),
                                     M_solid->mesh().getRefFE(), M_solid->dDof());
        M_dofStructureToSolid->update(M_solid->mesh(), 1,
                                      M_solid->mesh(), 1,
                                      0.);

        M_dofStructureToHarmonicExtension->setup(M_fluid->mesh().getRefFE(), M_fluid->harmonicExtension().dofMesh(),
//        M_dofStructureToHarmonicExtension->setup(M_fluid->mesh().getRefFE(), M_fluid->uDof(),
                                                 M_solid->mesh().getRefFE(), M_solid->dDof());
        M_dofStructureToHarmonicExtension->update(M_fluid->mesh(), 1,
                                                  M_solid->mesh(), 1,
                                                  0.0);

        M_dofHarmonicExtensionToFluid->setup(M_fluid->mesh().getRefFE(),M_fluid->uDof(),
                                             M_fluid->mesh().getRefFE(),M_fluid->uDof());
        M_dofHarmonicExtensionToFluid->update(M_fluid->mesh(), 1,
                                 M_fluid->mesh(), 1,
                                 0.0);

        M_dofStructureToReducedFluid->setup(M_fluid->mesh().getRefFE(), M_fluid->pDof(),
                                            M_solid->mesh().getRefFE(), M_solid->dDof());
        M_dofStructureToReducedFluid->update(M_fluid->mesh(), 1,
                                             M_solid->mesh(), 1,
                                             0.0);

        M_dofReducedFluidToStructure->setup(M_solid->mesh().getRefFE(), M_solid->dDof(),
                                            M_fluid->mesh().getRefFE(), M_fluid->uDof());
        M_dofReducedFluidToStructure->update(M_solid->mesh(), 1,
                                             M_fluid->mesh(), 1,
                                             0.0);
    }
}
//

void
FSIOperator::setDataFromGetPot( GetPot const& data_file )
{
    M_solverAztec.setOptionsFromGetPot(data_file,"jacobian/aztec");
    M_method  = data_file("problem/method" ,0);
    M_reducedFluid = data_file("problem/reducedFluid", 0);
    std::cout << "setdatafromgetpot: reducedFluid " << M_reducedFluid << std::endl;
    if (M_reducedLinFluid) M_reducedLinFluid->setLinearSolver(data_file);
}

//

void
FSIOperator::updateJacobian(Vector& /*sol*/,int /*iter*/)
{
}

//


Vector
FSIOperator::displacementOnInterface()
{

    Vector dispOnInterface(M_solid->disp().size());
    dispOnInterface = ZeroVector(dispOnInterface.size());

    FOR_EACH_INTERFACE_DOF( dispOnInterface[IDsolid - 1 + jDim*totalDofSolid] =
                            M_solid->disp()              [IDsolid - 1 + jDim*totalDofSolid]);

    std::cout << "max norm disp = " << norm_inf(dispOnInterface);
    std::cout << std::endl;
    std::cout << "l2  norm disp = " << norm_2(dispOnInterface);
    std::cout << std::endl;

    return dispOnInterface;
}


void
FSIOperator::transferOnInterface(const Vector      &_vec1,
                                 const BCHandler   &_BC,
                                 const std::string &_BCName,
                                 Vector            &_vec2)
{
    _vec2 = ZeroVector(_vec2.size());

    int iBC = _BC.getBCbyName(_BCName);

    BCBase const &BCInterface = _BC[(UInt) iBC];

    UInt nDofInterface = BCInterface.list_size();

//     std::cout << "nDofInterface = " << nDofInterface << std::endl;

    UInt nDim = BCInterface.numberOfComponents();

    UInt totalDof1 = _vec1.size()/ nDim;
    UInt totalDof2 = _vec2.size()/ nDim;

    for (UInt iBC = 1; iBC <= nDofInterface; ++iBC)
    {
        ID ID1 = BCInterface(iBC)->id();

        BCVectorInterface const *BCVInterface =
            static_cast <BCVectorInterface const *>
            (BCInterface.pointerToBCVector());

        ID ID2 = BCVInterface->
            dofInterface().getInterfaceDof(ID1);

        for (UInt jDim = 0; jDim < nDim; ++jDim)
        {
            _vec2[ID2 - 1 + jDim*totalDof2] =
                _vec1[ID1 - 1 + jDim*totalDof1];
        }
    }
}


//

void FSIOperator::setHarmonicExtensionVelToFluid(PhysVectUnknown<Vector> &vel,
                                                 UInt type)
{
    M_bcvHarmonicExtensionVelToFluid->setup(vel,
                                            M_fluid->uDof().numTotalDof(),
                                            M_dofHarmonicExtensionToFluid,
                                            type);
}

void FSIOperator::setDerHarmonicExtensionVelToFluid(PhysVectUnknown<Vector> &dvel,
                                                    UInt type)
{
    M_bcvDerHarmonicExtensionVelToFluid->setup(dvel,
                                               M_fluid->uDof().numTotalDof(),
                                               M_dofHarmonicExtensionToFluid,
                                               type);
}

void FSIOperator::setStructureDispToHarmonicExtension(PhysVectUnknown<Vector> &disp,
                                                      UInt type)
{
    M_bcvStructureDispToHarmonicExtension->setup(disp,
                                                 M_solid->dDof().numTotalDof(),
                                                 M_dofStructureToHarmonicExtension,
                                                 type);
}

void FSIOperator::setStructureDispToSolid(PhysVectUnknown<Vector> &disp,
                                          UInt type)
{
    M_bcvStructureDispToSolid->setup(disp,
                                     M_solid->dDof().numTotalDof(),
                                     M_dofStructureToSolid,
                                     type);
}

void FSIOperator::setDerStructureDispToSolid(PhysVectUnknown<Vector> &ddisp,
                                             UInt type)
{
    M_bcvDerStructureDispToSolid->setup(ddisp,
                                        M_solid->dDof().numTotalDof(),
                                        M_dofStructureToSolid,
                                        type);
}


void FSIOperator::setFluidLoadToStructure(Vector &load,
                                          UInt type)
{
    M_bcvFluidLoadToStructure->setup(load,
                                     M_fluid->uDof().numTotalDof(),
                                     M_dofFluidToStructure,
                                     type);
}

void FSIOperator::setDerFluidLoadToStructure(Vector &dload,
                                             UInt type)
{
    M_bcvDerFluidLoadToStructure->setup(dload,
                                        M_fluid->uDof().numTotalDof(),
                                        M_dofFluidToStructure,
                                        type);
}

void FSIOperator::setDerFluidLoadToFluid(Vector &dload,
                                         UInt type)
{
    M_bcvDerFluidLoadToFluid->setup(dload,
                                    M_fluid->uDof().numTotalDof(),
                                    M_dofHarmonicExtensionToFluid,
                                    type);
}

void FSIOperator::setDerReducedFluidLoadToStructure(Vector &dload,
                                                    UInt type)
{
    M_bcvDerReducedFluidLoadToStructure->setup(dload,
                                               M_fluid->uDof().numTotalDof(),
                                               M_dofReducedFluidToStructure,
                                               type);
}

void FSIOperator:: setDerStructureAccToReducedFluid(Vector &acc,
                                                    UInt type)
{

    M_bcvDerStructureAccToReducedFluid->setup(acc,
                                              M_solid->dDof().numTotalDof(),
                                              M_dofStructureToReducedFluid,
                                              type);
}
}
